/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.io.*;

import ij.IJ;
import ij.ImagePlus;
import ij.gui.GenericDialog;
import ij.io.*;
import ij.plugin.filter.Analyzer;
import ij.plugin.frame.Recorder;
import ij.text.*;
import ij.util.Tools;

public class CatalogedImageInfo {

	public static final String POSTFIX = "_cat.txt";
	public static final char SEP_CHAR = '\t';
	
	public ImagePlus imp;
	public String impDir = "";
	public String impName = "";
	public String impExt = "";
	public String impFormat = "";
	public String sample = "";
	public String muscle = "";
	public String experiment = "";
	public String probe = "";
	public String description = "";
	public String info = "";
	public double timeExp = 1.0;
	public double pixelDepth = 1.0;
	public double pixelHeight = 1.0;
	public double pixelWidth = 1.0;
	double[][] extHeaderValues;
	String[] extHeaderLabels;
	
	public CatalogedImageInfo(ImagePlus imp) {
		this.imp = imp;
		impDir = imp.getOriginalFileInfo().directory;
		impName = imp.getShortTitle();
		String impTitle = imp.getTitle();
		int index = impTitle.lastIndexOf(".");
		impExt = impTitle.substring(index,impTitle.length());
		int format = imp.getOriginalFileInfo().fileFormat;
		if (format == FileInfo.TIFF) impFormat = "TIFF";
		else if (format == FileInfo.RAW && impExt == ".dv") {
			if (impName.endsWith("_R3D")) impFormat = "DV-raw";
			else impFormat = "DV-deconv";
			String header = imp.getOriginalFileInfo().info;
			if (header != null) parseDVHeader(header);
		}
		info = imp.getCalibration().info;
		if (extHeaderValues != null) timeExp = getExposureTime();
		pixelDepth = imp.getCalibration().pixelDepth;
		pixelHeight = imp.getCalibration().pixelHeight;
		pixelWidth = imp.getCalibration().pixelWidth;
	}
	
/*	public CatalogedImageInfo(String dir, String fname, boolean exists) {
		imgdir = dir;
		imgname = fname;
		if (exists) load();
		else edit();
	}
*/	
	public void parseDVHeader(String header) {
		String [] values = Tools.split(header,"\n");
		int index = 0;
		while (!values[index].startsWith("ExtendedHeader")) index++;
		int headerEntries = 14;
		extHeaderValues = new double[imp.getImageStackSize()][headerEntries];
		extHeaderLabels = new String[headerEntries];
		int s0,s1,s2;
		for (int i=0; i<imp.getImageStackSize(); i++) {
			for (int j=0; j<headerEntries; j++) {
				s0 = values[index].indexOf("_");
				s1 = values[index].lastIndexOf(":");
				s2 = values[index].length();
				if (i==0) extHeaderLabels[j] = values[index++].substring(s0,s1);
				extHeaderValues[i][j] = Double.parseDouble(values[index++].substring(s1,s2));
			}
		}
		System.out.println();
	}
	
	public double getExposureTime() {
		int index = 0;
		while (extHeaderLabels[index] != "EXPOSURE_TIME") index++;
		return extHeaderValues[0][index];
	}
	
	public boolean isSaved() {
		File f = new File(getSavePath());
		return f.exists();
	}
	
	public void load() {
		String path = getSavePath();
		String[] values = null;
		
		IJ.showStatus("Opening: " + path);
		try {
			BufferedReader r = new BufferedReader(new FileReader(path));
			values = load2(r);
			r.close();
		}
		catch (Exception e) {
			IJ.error(e.getMessage());
		}
		IJ.showStatus("");
		if (values != null && values.length > 0) {
			parseValues(values);
		}
	}
	
	public String[] load2(BufferedReader in) throws IOException {
		String header = in.readLine();
		String data = in.readLine();
		return Tools.split(data,""+SEP_CHAR);
	}
	
	public void parseValues(String[] values) {
		int index = 1;
		impName = values[index++];
		impDir = values[index++];
		experiment = values[index++];
		sample = values[index++];
		muscle = values[index++];
		probe = values[index++];
		description = values[index++];
		info = values[index++];
		timeExp = Double.parseDouble(values[index++]);
		pixelDepth = Double.parseDouble(values[index++]);
		pixelHeight = Double.parseDouble(values[index++]);
		pixelWidth = Double.parseDouble(values[index++]);
	}

	public void edit() {
		GenericDialog gd = new GenericDialog("Catalog Image");
		gd.addStringField("Sample",sample,15);
		gd.addStringField("Muscle",muscle,30);
		gd.addStringField("Experiment",experiment,60);
		gd.addStringField("Probe",probe,15);
		gd.showDialog();
		sample = gd.getNextString();
		muscle = gd.getNextString();
		experiment = gd.getNextString();
		probe = gd.getNextString();
//		save();
	}
	
	public void save() {
		PrintWriter pw = null;
		try {
			FileOutputStream fos = new FileOutputStream(getSavePath());
			BufferedOutputStream bos = new BufferedOutputStream(fos);
			pw = new PrintWriter(bos);
		}
		catch (IOException e) {
			//IJ.write("" + e);
			return;
		}
		pw.println(getLabels());
		pw.println(getInfoString());
		pw.close();
		IJ.showStatus("");
	}
	
	public String getLabels() {
		StringBuffer sb = new StringBuffer();
		sb.append("Version" + SEP_CHAR);
		sb.append("Img FileName" + SEP_CHAR);
		sb.append("Img Directory" + SEP_CHAR);
		sb.append("Experiment" + SEP_CHAR);
		sb.append("Sample" + SEP_CHAR);
		sb.append("Muscle" + SEP_CHAR);
		sb.append("Probe" + SEP_CHAR);
		sb.append("Img Description" + SEP_CHAR);
		sb.append("Img Info" + SEP_CHAR);
		sb.append("Exposure Time" + SEP_CHAR);
		sb.append("Img Pixel Depth" + SEP_CHAR);
		sb.append("Img Pixel Height" + SEP_CHAR);
		sb.append("Img Pixel Width" + SEP_CHAR);
		return new String(sb);
	}
	
	public String getInfoString() {
		StringBuffer sb = new StringBuffer();
		sb.append(LSDDecon_Manager.VERSION + SEP_CHAR);
		sb.append(impName + SEP_CHAR);
		sb.append(impDir + SEP_CHAR);
		sb.append(experiment + SEP_CHAR);
		sb.append(sample + SEP_CHAR);
		sb.append(muscle + SEP_CHAR);
		sb.append(probe + SEP_CHAR);
		sb.append(description + SEP_CHAR);
		sb.append(info + SEP_CHAR);
		sb.append(timeExp + SEP_CHAR);
		sb.append(pixelDepth + SEP_CHAR);
		sb.append(pixelHeight + SEP_CHAR);
		sb.append(pixelWidth + SEP_CHAR);
		return new String(sb);
	}
	
	public String toString() {
		String s= this.getClass().toString() + ": dir= "+impDir
											 + "; name="+impName
											 + "; sample="+sample
											 + "; muscle="+muscle
											 + "; experiment="+experiment
											 + "; probe="+probe
											 ;
		return s;
	}

	public String confirm() {
		String s= "name= " + impName
				+ "\n" + "dir = "+impDir
				+ "\n" + "sample = "+sample
				+ "\n" + "muscle = "+muscle
				+ "\n" + "experiment = "+experiment
				+ "\n" + "probe = "+probe
				;
		return s;
	}
	
	public String getSavePath() {
		return (impDir + impName + POSTFIX);
	}
}
