/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.awt.geom.Point2D;

import ij.IJ;
import ij.gui.GenericDialog;

/** DDCurveFitter is a modified version of the CurveFitter class which exclusively
 *  handles the fitting of distributed deconvolution profiles to linescans.  It uses
 *  a modified version of the simplex method.  It has a custom dialog box for fitting
 *  options relavent to ddecon analysis and it acquires parameter values and function 
 *  evaluations from the ddecon plugin.
 *  
 *  Only the free DDecon parameters are passed to the fitter.  The original ddecon
 *  parameters are actually reparameterized when passed to the fitter such that the 
 *  original parameters can be constrained (i.e. have upper and/or lower bounds), but
 *  the passed parameters are unbounded and behave nicely in this method.
 *  
 *  As with many methods for optimizing non-linear models, it is possible that this 
 *  method does not find the global solution, however, based on experience with this
 *  method and this problem, it is very likely that this method will find the best fit
 *  if the model parameters are set correctly and it is started close enough to the 
 *  solution. 
 * 
 * @author Ryan Littlefield (ryanccd@u.washington.edu)
 * @see CurveFitter.class
 * @see DDecon_.class
 * @version 1.00c
 *
 */
public class DDCurveFitter {    
    public static final int IterFactor = 500;
    
    public static final String[] fitList = {"Simplex"};
    public static final char conv = '\u2297';
    public static final String function = "PSF() "+conv+" Model() "+conv+" Repeats()";
    
	public static final double MAX_ZP_DEVIATION = 5; // default maximum change of each zp during fitting, should be <= MAX_OVERHANG;
	public static final double MAX_ZW_DEVIATION = 0.4; // default maximum fractional change in Zw during fitting
	static final double MIN_TFL = 3.0;
	static final double MAX_TFL = 15.0;

    private static final double alpha = -1.0;	  // reflection coefficient
    private static final double beta = 0.5;	  // contraction coefficient
    private static final double gamma = 2.0;	  // expansion coefficient
    private static final double root2 = 1.414214; // square root of 2
    
    private double[] xData, yData;  // x,y data to fit
    private int numPoints;          // number of data points
    private int numParams;          // number of parametres
    private int numVertices;        // numParams+1 (includes sumLocalResiduaalsSqrd)
    private int worst;			// worst current parametre estimates
    private int nextWorst;		// 2nd worst current parametre estimates
    private int best;			// best current parametre estimates
    private double[][] simp; 		// the simplex (the last element of the array at each vertice is the sum of the square of the residuals)
    private double[] next;		// new vertex to be tested
    public int numIter;		// number of iterations so far
    public int maxIter; 	// maximum number of iterations per restart
    public int setRestarts; 	// number of times to restart simplex after first soln.
    private int restarts; 	// number of times to restart simplex after first soln.
    public double maxPosSearch = MAX_ZP_DEVIATION; // default search
    public double maxIntSearch = MAX_ZW_DEVIATION;
    private double maxTFlen = MAX_TFL;
    private double minTFlen = MIN_TFL;
    boolean[][] pStates = new boolean[3][];
//    private int maxOverhang = DDplotter.MAX_OVERHANG;
    public String info = "";
    public double maxError;     // maximum error tolerance
    private double maxY; // = 0;
    public boolean repeatDistWeight = false;

	double max_pos_search;
	double max_int_search;
	public double minTFL;
	public double maxTFL;

    
//    public DDProcessor dd;
//    public DDplotter ddp;
    public RepeatModel rm;
//    public DDinfo ddi;
    
    /** Construct a new CurveFitter. */
    public DDCurveFitter (RepeatModel rm, double[] lsInts) {
    	this.rm = rm;
        this.yData = lsInts;
        numPoints = yData.length;
        this.xData = new double [numPoints];
        for (int i=0; i<numPoints; i++) xData[i] = i;
        
        maxY = yData[getMax(yData)];
    }
    
    /** Construct a new CurveFitter. */
/*    public DDCurveFitter (DDProcessor dd) {
        this.dd = dd;
        this.xData = new double [dd.ddi.nDataPoints];
        for (int i=0; i<xData.length; i++) xData[i] = i;
        this.yData = dd.ddi.lsi.corrLSint;
        numPoints = yData.length;
        maxY = yData[getMax(yData)];
    }
*/    
    /** Construct a new CurveFitter. */
/*    public DDCurveFitter (double[] xData, double[] yData, DDProcessor dd) {
        this.xData = xData;
        this.yData = yData;
        this.dd = dd;
//        this.ddi = dd.ddi;
        numPoints = yData.length;
        for (int i = 0; i < numPoints; i++) if (yData[i]>maxY) maxY = yData[i];
    }
*/    
    /**  Perform curve fitting with the simplex method
     *          doFit(fitType) just does the fit
     *          doFit(fitType, true) pops up a dialog allowing control over simplex parameters
     *  	alpha is reflection coefficient  (-1)
     *  	beta is contraction coefficient (0.5)
     *  	gamma is expansion coefficient (2)
     */
    public boolean doFit() {
        return doFit(false);
    }
    
    public boolean doFit(boolean showSettings) {
    	initOptions();
    	if (showSettings) settingsDialog();
    	
    		max_pos_search = Math.min(maxPosSearch,maxPSearchValue());
    		max_int_search = maxIntSearch;
    		// setting min and max thin filament lengths limited by the current value
    		double currLen = rm.model.getParams()[rm.model.getLengthIndex()];
    		minTFL = Math.min(minTFlen,currLen);
    		maxTFL = Math.max(maxTFlen,currLen);
//    	dd.setFitOptions(maxPosSearch,maxIntSearch,minTFlen,maxTFlen);
            double[] params = getEncodedParams();
        initialize(params);
        restart(0);
        
        numIter = 0;
        boolean done = (maxIter==0); // if maxIter is zero, done is true and fit is not performed
        boolean fit = !done; // if done is true and fit is thus not performed, fit is false
        double rawE; // Raw Error (sum of the squares of the residuals)
        double[] center = new double[numParams];  // mean of simplex vertices

//	        double lastTime = System.currentTimeMillis();
        while (!done) {
            numIter++;
            rawE = simp[best][numParams];
            IJ.showStatus("Iteration=" + numIter + "...Normalized Error = "
               	 + IJ.d2s(100*Math.sqrt(rawE)/maxY,2) + " Raw Error = " + IJ.d2s(rawE));
//            	 + IJ.d2s(100*Math.sqrt(rawE/numPoints)/maxY,2) + " Raw Error = " + IJ.d2s(rawE));

            for (int i = 0; i < numParams; i++) center[i] = 0.0;
            // get mean "center" of vertices, excluding worst
            for (int i = 0; i < numVertices; i++)
                if (i != worst)
                    for (int j = 0; j < numParams; j++)
                        center[j] += simp[i][j];
            
            // Reflect worst vertex through centre
            for (int i = 0; i < numParams; i++) {
                center[i] /= numParams;
                next[i] = center[i] + alpha*(simp[worst][i] - center[i]);
            }
           	sumResiduals(next);
	            	
            // if it's better than the best...
            if (next[numParams] <= simp[best][numParams]) {
                newVertex();
                // try expanding it
                for (int i = 0; i < numParams; i++)
                    next[i] = center[i] + gamma * (simp[worst][i] - center[i]);
                sumResiduals(next);
                // if this is even better, keep it
                if (next[numParams] <= simp[worst][numParams])
                    newVertex();
            }
            
            // else if better than the 2nd worst keep it...
            else if (next[numParams] <= simp[nextWorst][numParams]) {
                newVertex();
            }
            
            // else try to make positive contraction of the worst
            else {
                for (int i = 0; i < numParams; i++)
                    next[i] = center[i] + beta*(simp[worst][i] - center[i]);
                sumResiduals(next);
                // if this is better than the second worst, keep it.
                if (next[numParams] <= simp[nextWorst][numParams]) {
                    newVertex();
                }
                // if all else fails, contract simplex in on best
                else {
                    for (int i = 0; i < numVertices; i++) {
                        if (i != best) {
                            for (int j = 0; j < numVertices; j++)
                                simp[i][j] = beta*(simp[i][j]+simp[best][j]);
                            sumResiduals(simp[i]);
                        }
                    }
                }
            }
            order();
            
            double rtol = 2 * Math.abs(simp[best][numParams] - simp[worst][numParams]) /
            (Math.abs(simp[best][numParams]) + Math.abs(simp[worst][numParams]) + 0.0000000001);
            
            if (numIter >= maxIter) done = true;
            else if (rtol < maxError) {
                //System.out.print(getResultString());
                restarts--;
                if (restarts < 0) {
                    done = true;
                }
                else {
                    restart(best);
                }
            }
        }
        return fit;
    }
    
    void initOptions() {
    	numParams = getNumParams();
    	maxPosSearch = maxPSearchValue()-0.1;
        maxIter = IterFactor * numParams * numParams;  // Where does this estimate come from?
        restarts = 5;
        maxError = 1e-9;
    }
        
    /** Pop up a dialog allowing control over simplex starting parameters */
    private void settingsDialog() {
        GenericDialog gd = new GenericDialog("DDecon Fitting Options", IJ.getInstance());
        gd.addMessage("Method: " + fitList[0] + "\n" +
                "Formula: " + function);
        gd.addNumericField("Maximum iterations:", maxIter, 0);
        gd.addNumericField("Number of restarts:", restarts, 0);
        gd.addNumericField("Position search distance:", maxPosSearch, 1);
        gd.addNumericField("Intensity search percentage:", (100*maxIntSearch), 0);
        gd.addNumericField("Min_TF Length:",minTFlen,0);
        gd.addNumericField("Max_TF Length:",maxTFlen,0);
        gd.addNumericField("Error tolerance [1*10^(-x)]:", -(Math.log(maxError)/Math.log(10)), 0);
//	        gd.addTextAreas();
        gd.addCheckbox("Weight by distance from repeat center:",repeatDistWeight);
        gd.addMessage("All parameters are limited to positive numbers.");
        gd.showDialog();
        if (gd.wasCanceled() || gd.invalidNumber()) {
            IJ.showStatus("Fit canceled.");
            maxIter = 0; // This will prevent the fit from occuring
            return;
        }
        maxIter = (int) gd.getNextNumber();
        setRestarts = (int) gd.getNextNumber();
        restarts = setRestarts;
        maxPosSearch = gd.getNextNumber();
        maxIntSearch = gd.getNextNumber()/100;
        minTFlen = gd.getNextNumber();
        maxTFlen = gd.getNextNumber();
        if (minTFlen>maxTFlen) {
        	// swap min and max TF lengths
        	double temp = minTFlen;
        	minTFlen = maxTFlen;
        	maxTFlen = temp;
        }
        maxError = Math.pow(10.0, -gd.getNextNumber());
        repeatDistWeight = gd.getNextBoolean();
    }
    
    /** Initialise the simplex
     */
    void initialize(double[] params) {
        // Calculate some things that might be useful for predicting parametres
        if (params.length != numParams) {
        	IJ.write("DDfitter.initialize: ERROR!!!! params.length="+params.length
        			+"; numParams="+numParams);
        }
        numParams = params.length;
        numVertices = numParams + 1;      // need 1 more vertice than parametres,
        simp = new double[numVertices][numVertices];
        next = new double[numVertices];
        
        for (int i=0;i<numParams; i++) simp[0][i] = params[i];
        simp[0][numParams] = 0.0;
    }
    
    /** Initialise the simplex
     */
/*    void initialize() {
        // Calculate some things that might be useful for predicting parametres
        double[] params = dd.getEncodedParams();
        if (params.length != numParams) {
        	System.out.println("DDfitter.initialize: ERROR!!!! params.length="+params.length
        			+"; numParams="+numParams);
        }
        numParams = params.length;
        numVertices = numParams + 1;      // need 1 more vertice than parametres,
        simp = new double[numVertices][numVertices];
        next = new double[numVertices];
        
        for (int i=0;i<numParams; i++) simp[0][i] = params[i];
        simp[0][numParams] = 0.0;
    }
*/    
    /** Restart the simplex at the nth vertex */
    void restart(int n) {
        // Copy nth vertice of simplex to first vertice
        for (int i = 0; i < numParams; i++) {
            simp[0][i] = simp[n][i];
        }
        sumResiduals(simp[0]);          // Get sum of residuals^2 for first vertex
        double[] step = getSteps();
//        double[] step = new double[numParams];
        for (int i = 0; i < numParams; i++) {
//       		step = getSteps();
            if (step[i] == 0.0)             // We can't have them all the same or we're going nowhere
                step[i] = 0.01;
        }

        // Some kind of factor for generating new vertices
        double[] p = new double[numParams];
        double[] q = new double[numParams];
        for (int i = 0; i < numParams; i++) {
            p[i] = step[i] * (Math.sqrt(numVertices) + numParams - 1.0)/(numParams * root2);
            q[i] = step[i] * (Math.sqrt(numVertices) - 1.0)/(numParams * root2);
        }

        // Create the other simplex vertices by modifing previous one.
        for (int i = 1; i < numVertices; i++) {
            for (int j = 0; j < numParams; j++) {
                simp[i][j] = simp[i-1][j] + q[j]; // all the parameters are modified by q
            }
            simp[i][i-1] = simp[i][i-1] + p[i-1]; // one parameter for each vertex is modified by p
            sumResiduals(simp[i]);
        }

        // Initialise current lowest/highest parametre estimates to simplex 1
        best = 0;
        worst = 0;
        nextWorst = 0;
        order();
    }
        
    // Display simplex [Iteration: s0(p1, p2....), s1(),....] in ImageJ window
    void showSimplex(int iter) {
        ij.IJ.write("" + iter);
        for (int i = 0; i < numVertices; i++) {
        	double [] overtex = decodeParams(simp[i]);
            String s = "";
            for (int j=0; j < overtex.length; j++)
                s += "  "+ ij.IJ.d2s(overtex[j], 4);
            ij.IJ.write(s);
        }
	}
        
    /** Get number of parameters */
    public int getNumParams() {
    	return rm.getNumFreeParams();
    	}
        
    /** Get the set of parameter values from the best corner of the simplex */
    public double[] getParams() {
        order();
        return simp[best];
    }
    
    /** Returns residuals array ie. differences between data and curve. */
/*    public double[] getResiduals() {
    	return dd.getPFresiduals();
    }
*/    
    /* Last "parametre" at each vertex of simplex is sum of residuals
     * for the curve described by that vertex
     */
    public double getSumResidualsSqr() {
        double sumResidualsSqr = (getParams())[getNumParams()];
        return sumResidualsSqr;
    }
    
    public double getChiSquared() {
	/* This assumes that signal dependent shot-noise proportional to sqrt of model intensity dominates
	 * the sigma for each data point and that dark noise any readout noise are absent.  This also 
	 * does not give the appropriate scale to the shot-noise, so the chi-square value is only relative.
	 * 
	 * TO DO add scaling term for shot-noise specific for camera conditions
	 * TO DO add dark-noise and readout noise components specific for camera conditions
	 * TO DO offset intensities with the dark field average intensity specific for camera conditions
	 */
    	
    	double beta = 0.315; // calibrated from exposure series
    	double varDN = 3; // estimated
    	double i0 = 90; // estimated
    	double varI;
    	
    	double[] r = new double[numPoints];
    	for (int i=0; i<numPoints; i++) {
    		r[i] = rm.calcInts[i] - yData[i];
    	}
//    	double[] r = getResiduals();
    	double chi2 = 0;
    	for (int i=0; i<yData.length; i++) {
//	    		chi2 += r[i]*r[i]/yData[i]; // sum of residuals squared weighted by the observed intensity
        	varI = yData[i];
//	        	varI = varDN + beta*(yData[i]-i0);
    		chi2 += r[i]*r[i]/varI; // sum of residuals squared weighted by the total variance
    	}
    	return chi2;
    }
    
    /**  SD = sqrt(sum of residuals squared / number of points)
     */
    public double getSD() {
        double sd = Math.sqrt(getSumResidualsSqr() / numPoints);
        return sd;
    }
    
    /** Returns R^2, where 1.0 is best.
    <pre>
     r^2 = 1 - SSE/SSD
     
     where:	 SSE = sum of the squares of the errors
                 SSD = sum of the squares of the deviations about the mean.
    </pre>
    */
    public double getRSquared() {
        double sumY = 0.0;
        for (int i=0; i<numPoints; i++) sumY += yData[i];
        double mean = sumY/numPoints;
        double sumMeanDiffSqr = 0.0;
        for (int i=0; i<numPoints; i++)
            sumMeanDiffSqr += sqr(yData[i]-mean);
        double rSquared = 0.0;
        if (sumMeanDiffSqr>0.0)
            rSquared = 1.0 - getSumResidualsSqr()/sumMeanDiffSqr;
        return rSquared;
    }

    /**  Get a measure of "goodness of fit" where 1.0 is best. */
    public double getFitGoodness() {
        double sumY = 0.0;
        for (int i = 0; i < numPoints; i++) sumY += yData[i];
        double mean = sumY / numPoints;
        double sumMeanDiffSqr = 0.0;
        int degreesOfFreedom = numPoints - getNumParams();
        double fitGoodness = 0.0;
        for (int i = 0; i < numPoints; i++) {
            sumMeanDiffSqr += sqr(yData[i] - mean);
        }
        if (sumMeanDiffSqr > 0.0 && degreesOfFreedom != 0)
            fitGoodness = 1.0 - (getSumResidualsSqr() / degreesOfFreedom) * ((numParams) / sumMeanDiffSqr);
        
        return fitGoodness;
    }
    
    /** Get a string description of the curve fitting results
     * for easy output.
     */
    public String getResultString() {
        StringBuffer results = new StringBuffer("\nNumber of iterations: " + getIterations() +
        "\nMaximum number of iterations: " + getMaxIterations() +
        "\nSum of residuals squared: " + IJ.d2s(getSumResidualsSqr(),4) +
        "\nStandard deviation: " + IJ.d2s(getSD(),4) +
        "\nR^2: " + IJ.d2s(getRSquared(),4) +
        "\nParameters:");
        char pChar = 'a';
        double[] pVal = decodeParams(getParams());
        for (int i = 0; i < pVal.length-1; i++) {
            results.append("\n  " + pChar + " = " + IJ.d2s(pVal[i],4));
            pChar++;
        }
        int dof = numPoints - numParams;
//	        IJ.write("DDCurveFitter.getResultsString: estimated varience= "+IJ.d2s(getSumResidualsSqr()/dof,4));
        return results.toString();
    }
    
    /** Passes values from the fit to the DDinfo object for saving */
/*    public void setInfo(LSDDecon_Manager manager) {
//	    	IJ.write("Setting fit info");
    	manager.fitMethod = DDCurveFitter.fitList[0];//"Simplex";
    	manager.maxIter = this.maxIter;
    	manager.restarts = this.setRestarts;
    	manager.errTol = this.maxError;
    	manager.infoFit = this.info;
    	manager.max_int_search = this.maxIntSearch;
    	manager.max_overhang = repeatModel.MAX_OVERHANG;
    	manager.max_pos_search = this.maxPosSearch;
    	manager.numIter = this.numIter;
    	manager.goodness = this.getFitGoodness();
    	manager.ssr = this.getSumResidualsSqr();
    	manager.chi2 = this.getChiSquared();
    }
*/        
    double sqr(double d) { return d * d; }
    
    /** Adds sum of square of residuals to end of array of parameters */
    void sumResiduals (double[] x) {
        x[numParams] = 0.0;
        double[] feval = eval(x);
    	double beta = 0.315; // calibrated from exposure series
    	double varDN = 3; // estimated
    	double i0 = 90; // estimated
    	double varI;
    	double[] weights = new double[numPoints];
    	double sumWeight = 0;
    	for (int i = 0; i < numPoints; i++) {
    		weights[i] = 1.0;
//    		weights[i] += rm.getDistToNearRepeat(i); // Experiment for non-uniform fitting
    		if (repeatDistWeight) weights[i] += sqr(sqr(sqr(sqr(getDistToNearRepeat(x,i))))); // Experiment for non-uniform fitting
//    		System.out.println("Fitter.sumResiduals: weights["+i+"]="+weights[i]);
    		sumWeight += weights[i];
    	}
        for (int i = 0; i < numPoints; i++) {
        	
//	 Sum of residuals squared (Raw Error)
        	x[numParams] += weights[i]*sqr(feval[i]-yData[i]);
        	
//	 Chi-squared statistic        	
//	        	varI = yData[i];
//	        	varI = varDN + beta*(yData[i]-i0);
//	            x[numParams] += (feval[i]-yData[i])*(feval[i]-yData[i])/varI;
//	            x[numParams] += (feval[i]-yData[i])*(feval[i]-yData[i])/yData[i];
        }
        x[numParams] /= sumWeight;
    }

	public double getDistToNearRepeat(double[] params, int position) {
		double [] oparams = decodeParams(params);
		int nmp = rm.model.getParamNumber();
		double dist = Math.abs(oparams[nmp+0] - position);
		for (int i=1; i<rm.repts.length; i++) {
			dist = Math.min(dist,Math.abs(oparams[nmp+i] - position));
		}
		return dist;
	}
	
    /** Keep the "next" vertex */
    void newVertex() {
        for (int i = 0; i < numVertices; i++)
            simp[worst][i] = next[i];
    }
    
    /** Find the worst, nextWorst and best current set of parameter estimates */
    void order() {
        for (int i = 0; i < numVertices; i++) {
            if (simp[i][numParams] < simp[best][numParams])	best = i;
            if (simp[i][numParams] > simp[worst][numParams]) worst = i;
        }
        nextWorst = best;
        for (int i = 0; i < numVertices; i++) {
            if (i != worst) {
                if (simp[i][numParams] > simp[nextWorst][numParams]) nextWorst = i;
            }
        }
    }

    /** Get number of iterations performed */
    public int getIterations() {
        return numIter;
    }
    
    /** Get maximum number of iterations allowed */
    public int getMaxIterations() {
        return maxIter;
    }
    
    /** Set maximum number of iterations allowed */
    public void setMaxIterations(int x) {
        maxIter = x;
    }
    
    /** Get number of simplex restarts to do */
    public int getRestarts() {
        return restarts;
    }
    
    /** Get number of simplex restarts initially set to do */
    public int getSetRestarts() {
        return setRestarts;
    }
    
    /** Set number of simplex restarts to do */
    public void setRestarts(int x) {
        restarts = x;
    }

    /**
     * Gets index of highest value in an array.
     * 
     * @param              Double array.
     * @return             Index of highest value.
     */
    public static int getMax(double[] array) {
        double max = array[0];
        int index = 0;
        for(int i = 1; i < array.length; i++) {
            if(max < array[i]) {
            	max = array[i];
            	index = i;
            }
        }
        return index;
    }

	/** Returns the maximum value for the repeat position search from the current positions that
	 *  prevents the possibility of the repeats extending past the 'overhang', and prevents swapping
	 *  of two adjacent repeats. */
	double maxPSearchValue() {

		// To prevent Z line swapping, maxValue is limited by the minimum half-sarcomere length.
		double[] hsl = new double [rm.repts.length-1];
		double maxValue = rm.ddLength + 2*RepeatModel.MAX_OVERHANG;
		for (int i=0;i<rm.repts.length-1;i++) {
			hsl[i] = (rm.repts[i+1].x - rm.repts[i].x)/2;
			if (hsl[i]<maxValue) maxValue = hsl[i];
		}
		
		// To prevent moving the first Z line out of bounds, maxValue is limited by its starting position
		double v = rm.repts[0].x + RepeatModel.MAX_OVERHANG;
		if (v < maxValue) maxValue = v;
		
		// To prevent moving the last Z line out of bounds, maxValue is limited by its starting position
		v = (rm.ddLength + RepeatModel.MAX_OVERHANG) - rm.repts[rm.repts.length-1].x;
		if (v < maxValue) maxValue = v;

		return maxValue;
	}
	
	/** Returns the bounded reparameterized parameter values which behave nicely in the simplex fitting method.
	 *  Positive values are enforced by passing the sqrt of the original value: include psfw, tfuni, tfpend.
	 *  A range of values is enforced by specifying a search center and radius around which the value can vary:
	 *  includes tflen, zp, and zw.
	 * @see decodeParams
	 * */
	public double[] getEncodedParams() {
		int nfp = rm.getNumFreeParams();
		double[] params = new double [nfp];
		int index = 0;
		double[] boundsLo = rm.model.getBoundsLo();
		double[] boundsHi = rm.model.getBoundsHi();
		boundsLo[rm.model.getLengthIndex()] = minTFL; // XXX Bad code, but hopefully it works for now
		boundsHi[rm.model.getLengthIndex()] = maxTFL; // XXX Bad code, but hopefully it works for now
		double[] p = rm.model.getParams();
		boolean[] s = rm.model.getParamStates();
		int l = p.length;
//		for (int i=0; i<l; i++) {
//			System.out.println("DDCurveFitter.getEncodedParams(): p="+p[i]+"; s="+s[i]+"; bL="+boundsLo[i]+"; bH="+boundsHi[i]);
//		}
		double[] encoded = rm.model.encodeParamValues(rm.model.getParams(),rm.model.getParamStates(),boundsLo,boundsHi);
		for (int i=0; i<encoded.length; i++) {
			params[i] = encoded[i];
//			System.out.println("DDCurveFitter.getEncodedParams(): params[i]="+params[i]+"; encoded[i]="+encoded[i]+"; i="+i);
			index++;
		}

		// for Z positions, pass angle parameters to limit their distances within a certain radius from the starting positions
		for (int i=0; i<rm.repts.length; i++) {
			if (!rm.reptFixStates[0][i]) params[index++] = 0; //zp[i];
		}

		// for Z weights, pass angle parameters to limit their intensities within a set percentage from the starting ones
		for (int i=0; i<rm.repts.length; i++) {
			if (!rm.reptFixStates[1][i]) params[index++] = 0; //zw[i]/fitScale;
		}

		return params;
	}
	
	/** Returns the original parameters upon receiving the bounded reparameterized passed parameters.
	 * @see getEncodedParams() */
	public double[] decodeParams(double [] pparams) {
		int nmp = rm.model.getParamNumber();
		int np = nmp + rm.repts.length*2;
		double [] oparams = new double [np+1];
		int nfp = pparams.length-1;
		int index = 0;
//		for (int i=0; i<pparams.length; i++){
//			System.out.println("DDCurveFitter.decodeParams: pparams[i]="+pparams[i]+"; i="+i);
//		}
		// Pass them to the model for decoding
		double[] mParams = new double [nmp];
		for (int i=0; i<nmp; i++) {
//			if (!this.pStates[0][i]) {
			if (!rm.model.getParamStates()[i]) {
				mParams[i] = pparams[index]; // XXX Something about this is not right for variLengthmodel when focus param is fixed
//				System.out.println("DDCurveFitter.decodeParams: pparams[index]="+pparams[index]+"; index="+index);
//				System.out.println("DDCurveFitter.decodeParams: mParams[i]="+mParams[i]+"; i="+i);
				index++;
			}
			else
				mParams[i] = rm.model.getParams()[i];
//				System.out.println("DDecon_.decodeParams: i="+i+"; mP="+mParams[i]+"; pP="+pparams[i]+"; mP="+this.modParams[i]);
		}
		

		double[] decodedModelParams;
		double[] boundsLo = rm.model.getBoundsLo();
		double[] boundsHi = rm.model.getBoundsHi();
		boundsLo[rm.model.getLengthIndex()] = minTFL; // XXX Bad code, but hopefully it works for now
		boundsHi[rm.model.getLengthIndex()] = maxTFL; // XXX Bad code, but hopefully it works for now
		decodedModelParams = rm.model.decodeParamValues(mParams,rm.model.getParamStates(),boundsLo,boundsHi);
		
		// Now assign all the decoded model params to oparams
		for (int i=0; i<nmp; i++) oparams[i] = decodedModelParams[i];
		
		for (int i=0; i<rm.repts.length; i++) {
			if (!rm.reptFixStates[0][i]) {
				oparams[nmp+i] = rm.repts[i].x + Math.sin(pparams[index++])*max_pos_search;
			}
			else oparams[nmp+i] = rm.repts[i].x;
		}
		for (int i=0; i<rm.repts.length; i++) {
			if (!rm.reptFixStates[1][i]) {
				oparams[nmp+i+rm.repts.length] = rm.repts[i].y*(1 + Math.sin(pparams[index++])*max_int_search);
			}
			else oparams[nmp+rm.repts.length+i] = rm.repts[i].y;
		}
		oparams[np] = pparams[nfp]; // this is the error value

		return oparams;
	}
	
	/** Returns the initial steps for each parameter to construct the simplex for fitting*/
	public double[] getSteps() {
		int nfp = rm.getNumFreeParams();
		int nmp = rm.model.getParamNumber();
		int nz = rm.repts.length;//(np-nmp)/2;
		int np = nmp + 2*nz; //pStates[0].length+2*pStates[1].length;
		double[] step = new double [np];
		double[] modelStep = rm.model.getSimplexSteps();
//			System.out.println("DDecon_.getSteps: cModel="+cModel.getTitle()+"; step.l="+step.length+
//					"; modelStep.l="+modelStep.length+"; nmp="+nmp);
		for (int i=0; i<nmp; i++) {
			step[i] = modelStep[i];
		}
		for (int i=0; i<nz; i++) {
			step[nmp+i] = 0.05;
			step[nmp+nz+i] = 0.05;
		}
		
		// select only the steps for the free parameters that were passed
		double[] stepf = new double [nfp];
		int index = 0;
		boolean[][] pStates = new boolean[3][];
		pStates[0] = rm.model.getParamStates();
		pStates[1] = rm.reptFixStates[0];
		pStates[2] = rm.reptFixStates[1];
		for (int i=0; i<pStates.length; i++)
			for (int j=0; j<pStates[i].length; j++)
				if (!pStates[i][j]) stepf[index++] = step[i];
		return stepf;
	}
	
	public double[] eval(double [] params) {
		// reparameterize passed params into original params
		double [] oparams = decodeParams(params);
		return rm.eval(oparams);
		
/*		
//		int np = oparams.length-1;
		int nmp = rm.model.getParamNumber();
		int nz = rm.repts.length;//(np - nmp)/2;
		
		double[] mParams = new double [nmp];
		for (int i=0; i<nmp; i++) {
			mParams[i] = oparams[i];
		}
		Point2D.Double[] repts = new Point2D.Double [nz];
		for (int i=0; i<nz; i++) {
			repts[i] = new Point2D.Double(oparams[nmp+i],oparams[nmp+i+nz]);
		}

		// Caluclate functions
		int min;
		min = Math.max((int)(mParams[0]*6),repeatModel.MIN_PSF_LENGTH);
		Point2D.Double[] d1 = dd.makePSF(mParams[0],min);
//			System.out.println("DDecon_.eval: cModel.getLengthIndex()="+cModel.getLengthIndex());
//			System.out.println("DDecon_.eval: mParams[index]="+mParams[cModel.getLengthIndex()]);
//			System.out.println("DDecon_.eval: MIN_TFD_LENGTH="+MIN_TFD_LENGTH);
		min = (int)Math.max(repeatModel.MIN_TFD_LENGTH,2*mParams[rm.model.getLengthIndex()]+7);
//			System.out.println("DDecon_.eval: min="+min);
		
		Point2D.Double[] d2 = rm.model.makePixels(mParams,min);
		Point2D.Double[] c1 = rm.convolve(d1,d2);
		Point2D.Double[] d3 = dd.makeZD(repts,rm.ddLength);
//		Point2D.Double[] d3 = makeZD(zdPts,this.pfLen);
		Point2D.Double[] c2raw = rm.convolve(c1,d3);
		Point2D.Double[] c2 = rm.truncate(c2raw,0,rm.ddLength);
//		Point2D.Double[] c2 = truncate(c2raw,0,this.pfLen);
		double[] output = new double[numPoints];
		for (int i=0; i<numPoints; i++) {
			output[i] = c2[i].y;
		}
		return output;
*/
	}
	
}
