/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import ij.IJ;
import ij.ImagePlus;
import ij.Prefs;
import ij.gui.ImageCanvas;
import ij.gui.Plot;
import ij.gui.PlotWindow;
import ij.process.ImageProcessor;
import ij.util.Tools;

import java.awt.Checkbox;
import java.awt.Color;
import java.awt.Label;
import java.awt.Panel;
import java.awt.TextField;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.geom.Point2D;

public class RepeatModel {

	public static int MAX_OVERHANG = (int)Prefs.get("LSDDecon.maxOverhang",5);
	public static final int MIN_PSF_LENGTH = 31;
	public static final int MIN_TFD_LENGTH = 31;
	public static final double REPEAT_SPACING = 18; // Average sarcomere length in pixels
	public static final int MAX_NUMBER_REPEATS = 20;

	static boolean[] variAll = {false, false, false, false};
	static boolean[] variEnd = {false, false, true, false};
	static boolean[] variUni = {false, false, false, true};
	static boolean[] variPSF = {false, true, true, true};
	static UniformModel phall = new UniformModel("Phalloidin 2"
			, "Uniform, continuous staining along the free portions of the " +
			"thin filaments with independent Z line staining (2 parameters)"
			, new double[] {1.6,8.0,0.4});
	static UniformPendModel allFree3 = new UniformPendModel("All Free3"
			, "Classic 3 parameter model, all 3 parameters free"
			, new double[] {1.6,8.0,0.0,0.0}, variAll);
	static UniformPendModel aact3a = new UniformPendModel("a-Act3 no BG"
			, "Classic 3 parameter model for alpha-actinin, no background"
			, new double[] {1.6,8.0,0.0,0.0}, variPSF);
	static UniformPendModel aact3b = new UniformPendModel("a-Act3 with BG"
			, "Classic 3 parameter model for alpha-actinin with background"
			, new double[] {1.6,8.0,0.05,0.0}, variUni);
	static UniformPendModel phall3a = new UniformPendModel("Phalloidin3"
			, "Classic 3 parameter model for phalloidin"
			, new double[] {1.6,8.0,0.4,0.0}, variUni);
	static UniformPendModel tmod3a = new UniformPendModel("Tmod3"
			, "Classic 3 parameter model for Tmod"
			, new double[] {1.6,8.0,1.0,5.0}, variEnd);
	static UniformPendModel actin3a = new UniformPendModel("Early Labeled Actin3"
			, "Classic 3 parameter model for Short-term Labeled Actin Incorporation"
			, new double[] {2.0,6.0,0.0,0.5}, variEnd);
	static UniformPendModel actin3b = new UniformPendModel("Late Labeled Actin3"
			, "Classic 3 parameter model for Long-term Labeled Actin Incorporation"
			, new double[] {2.0,7.5,0.4,0.0}, variUni);
	static VariLengthModel phall4a = new VariLengthModel("Non-uni Phall4"
			, "4 parameter model for phalloidin staining with gaps near Z line"
			, new double[] {1.6,8.0,4.0,1.5,0}, new boolean[] {false,false,false,false,true});
	static SModel[] Models = {allFree3,aact3a,aact3b,phall3a,tmod3a,actin3a,actin3b,phall4a};
	static int dModelIndex = 4;
	int cModIndex = dModelIndex; // current model index

	
	
	public DDwindow ddwin;
	public PSFModel psf = new PSFModel("Gaussian1", "Original PSF Model", new double[] {1.6}, new boolean[] {false});;
	public SModel model = (SModel)Models[(int)Prefs.get("LSDDecon.defModIndex",3)].clone();
	public Point2D.Double[] repts;
	public boolean[][] reptFixStates = new boolean[2][];
	public int ddLength;
	public Point2D.Double[] modelProfilePixels;
	public double[] calcInts;
	public double[] xvals;
	public Panel[] panControls;
	public Label[] labRepts;
	public TextField[] tfReptPositions;
	public TextField[] tfReptWeights;
	public Checkbox[] cbReptPositions;
	public Checkbox[] cbReptWeights;
	
	
	public RepeatModel(int ddLength) {
		this(null, null, new Point2D.Double[0], new boolean[2][0], ddLength);
	}
	
	public RepeatModel(PSFModel psf, SModel model, Point2D.Double[] repts, boolean[][] fixStates, int ddLength) {
		if (psf!=null) this.psf = psf;
		if (model!=null) this.model = model;
		this.model.psf = psf;
		this.repts = repts;
		this.reptFixStates = fixStates;
		this.ddLength = ddLength;
		xvals = new double[ddLength];
		for (int i=0; i<ddLength; i++) {
			xvals[i] = i;
		}

		validateReptsAndFixStates();

		modelProfilePixels = calcModelInts(this.model.getParams());
		calcInts = calcProfileInts(modelProfilePixels,this.repts);
	}
	
	public void validateReptsAndFixStates() {
		int reps = repts.length;
		double aveSL;
		double offset;
		if (reps==0) {
//			System.out.println("RM.validate 0: ddLength="+ddLength+"; reps="+reps);
			if (reptFixStates.length==2 && reptFixStates[0].length==reptFixStates[1].length &&
					reptFixStates[0].length!=0) {
				reps = reptFixStates[0].length;
				aveSL = reps>1 ? ddLength/(reps-1) : 1;
				offset = reps>1 ? 0 : ddLength/2;
//				System.out.println("RM.validate 1: ddLength="+ddLength+"; reps="+reps);
			} else {
				aveSL = Prefs.get("LSDDecon.aveSL",22);
				reps = (int)Math.ceil((ddLength + 2*MAX_OVERHANG)/aveSL); // Guess at number of repeats
				offset = reps>1 ? (ddLength - aveSL*(reps-1))/2 : ddLength/2; // Guess at offset value
				reptFixStates = new boolean[2][reps];
				for (int i=0; i<reps; i++) {
					reptFixStates[0][i] = false;
					reptFixStates[1][i] = false;
				}
				IJ.showStatus("Guessed "+reps+" repeating profiles with average distance = "+ aveSL+" pixels.");
//				System.out.println("RM.validate 2: ddLength="+ddLength+"; reps="+reps);
			}
			repts = new Point2D.Double[reps];
			for (int i=0; i<reps; i++) repts[i] = new Point2D.Double(aveSL*i+offset,1000);
		}
		else {
			boolean isValid = false;
			for (int i=0; i<reps; i++) if (repts[i].x!=0) isValid = true;
			if (!isValid) for (int i=0; i<reps; i++) repts[i].x = reps>1 ? ddLength/(reps-1)*i : ddLength/2;
			isValid = false;
			for (int i=0; i<reps; i++) if (repts[i].y>0) isValid = true;
			if (!isValid) for (int i=0; i<reps; i++) repts[i].y = 1000;
			if (reptFixStates.length!=2 || reptFixStates[0].length!=reps || reptFixStates[1].length!=reps) {
				reptFixStates = new boolean[2][reps];
				for (int i=0; i<reps; i++) {
					reptFixStates[0][i] = false;
					reptFixStates[1][i] = false;
				}
			}
		}
	}
	
	public void updateModel(int index) {
		cModIndex = index;
		SModel m = (SModel)Models[cModIndex].clone();
		Prefs.set("LSDDecon.defModIndex",cModIndex);
		model = m;
		model.psf = this.psf;
	}
	
	public Point2D.Double[] calcModelInts(double[] modelParams) {
		Point2D.Double[] psfPixels = psf.makePixels(modelParams[0]);
		Point2D.Double[] modelDistributionPixels = model.makePixels(modelParams);
		return convolve(psfPixels,modelDistributionPixels);
	}
	
	public double[] calcProfileInts(Point2D.Double[] modelProfilePixels, Point2D.Double[] repts) {
		int pLength = ddLength+2*MAX_OVERHANG;
		Point2D.Double[] repeatDistributionPixels = new Point2D.Double [pLength];
		for (int i=0; i<pLength; i++) {
			repeatDistributionPixels[i] = new Point2D.Double(i-MAX_OVERHANG,0);
		}

		int index = 0; // index of repeats
		double dx; // distance
		for (int i=0; (i<pLength-1 && index<repts.length); i++) {
			dx = repts[index].x-repeatDistributionPixels[i].x;
			if (dx<1) {
				repeatDistributionPixels[i].y = repts[index].y*(1-dx);
				repeatDistributionPixels[++i].y = repts[index].y*dx;
				index++;
			}
		}
		return truncate(convolve(modelProfilePixels,repeatDistributionPixels),0,ddLength);
	}
	
	public double getDistToNearRepeat(int position) {
		double dist = Math.abs(repts[0].x - (position+1));
		for (int i=1; i<repts.length; i++) {
			dist = Math.min(dist,Math.abs(repts[i].x - (position+1)));
		}
		return dist;
	}
	
	public double[] eval(double [] oparams) {
		int nmp = model.getParamNumber();
		int nz = repts.length;
		
		double[] mParams = new double [nmp];
		for (int i=0; i<nmp; i++) {
			mParams[i] = oparams[i];
		}
		Point2D.Double[] repts = new Point2D.Double [nz];
		for (int i=0; i<nz; i++) {
			repts[i] = new Point2D.Double(oparams[nmp+i],oparams[nmp+i+nz]);
		}

		// Caluclate functions
		Point2D.Double[] modelProfilePixels = calcModelInts(mParams);
		return calcProfileInts(modelProfilePixels,repts);
	}
	
	public Point2D.Double[] convolve(Point2D.Double[] f1, Point2D.Double[] f2) {
		int len1 = f1.length;
		int len2 = f2.length;
		if (len1>len2) return convolve(f2,f1);	// swap f1 and f2 around
		Point2D.Double[] g = new Point2D.Double [len1+len2-1];	// len1 <= len2
		double x = f1[0].x + f2[0].x;
		for (int i=0; i<g.length; i++) { // add up everything for each pixel in g
			g[i] = new Point2D.Double(x+i,0);
			for (int m=0; m<len1; m++) { // go through each pixel in smaller f
				int n = m - (len1 - 1 - i); // corresponding index in longer f
				if (n>=0 && n<len2) { // if this is in range...
					g[i].y += f1[m].y*f2[n].y; // add the product of the intensities
				}
			}
		}
		return g;
	}
	
	/** Returns only the points whose X values are >min and <=max */
	public double[] truncate(Point2D.Double[] in, int min, int max) {
		double[] out = new double [max-min];
		int index = 0;
		while (in[index].x<min) {index++;}
		int start = index;
		for (int i=0; i<out.length; i++) {
			out[i] = in[i+start].y;
		}
		return out;
	}
	
	public void setParamStates() {
		
	}
	
	public int getNumFreeParams() {
		boolean[][] pStates = new boolean[3][];
		pStates[0] = model.getParamStates();
//		pStates[0][0] = psf.pState;
		pStates[1] = reptFixStates[0];
		pStates[2] = reptFixStates[1];
		if (pStates==null) return 0;
		int index = 0;
		for (int i=0; i<pStates.length; i++)
			for (int j=0; j<pStates[i].length; j++)
				if (!pStates[i][j]) index++;
		return index;
	}
	
	public void makePanels(Panel panel, ActionListener aListener, ItemListener iListener) {
		panControls = new Panel[repts.length];
		labRepts = new Label[repts.length];
		tfReptPositions = new TextField[repts.length];
		tfReptWeights = new TextField[repts.length];
		cbReptPositions = new Checkbox[repts.length];
		cbReptWeights = new Checkbox[repts.length];
		for (int i=0; i<repts.length; i++) {
			makeRepeatControlPanel(i, aListener, iListener);
			labRepts[i].setText("Z"+(i+1));
			tfReptPositions[i].setText(IJ.d2s(repts[i].x,2));
			tfReptWeights[i].setText(IJ.d2s(repts[i].y,2));
			cbReptPositions[i].setState(reptFixStates[0][i]);
			cbReptWeights[i].setState(reptFixStates[1][i]);
			panControls[i].setBounds(140+70*i,0,70,100);
			panel.add(panControls[i]);
		}
	}
	
	public void makeRepeatControlPanel(int i, ActionListener aListener, ItemListener iListener) {
		panControls[i] = new Panel();
		labRepts[i] = new Label();
		tfReptPositions[i] = new TextField(3);
		cbReptPositions[i] = new Checkbox();
		tfReptWeights[i] = new TextField(3);
		cbReptWeights[i] = new Checkbox();
		
		panControls[i].setLayout(null);
		
		labRepts[i].setBounds(25,0,30,20);
		panControls[i].add(labRepts[i]);
		
		tfReptPositions[i].setBounds(5,20,60,20);
		tfReptPositions[i].addActionListener(aListener);
		panControls[i].add(tfReptPositions[i]);

		cbReptPositions[i].setBounds(25,40,30,20);
		cbReptPositions[i].addItemListener(iListener);
		panControls[i].add(cbReptPositions[i]);

		tfReptWeights[i].setBounds(5,60,60,20);
		tfReptWeights[i].addActionListener(aListener);
		panControls[i].add(tfReptWeights[i]);

		cbReptWeights[i].setBounds(25,80,30,20);
		cbReptWeights[i].addItemListener(iListener);
		panControls[i].add(cbReptWeights[i]);
	}
	
}
