/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.awt.geom.*;

public class UniformPendModel extends SModel {
	static final double BEndInt = 1.0;
	
	public UniformPendModel(String title, String desc, double[] dParams, boolean[] dStates) {
		super(title,desc,dParams,dStates);
		setParamNames(new String[] {"Focus","Length","Uniform","P-end"});
//		setParams(dParams);
//		setParamStates(states);
		setIsBoundLo(new boolean[] {true,true,true,true});
		setIsBoundHi(new boolean[] {false,true,false,false});
		setBoundsLo(new double[] {0.0,3.0,0.0,0.0});
		setBoundsHi(new double[] {Double.MAX_VALUE,15.0,Double.MAX_VALUE,Double.MAX_VALUE});
		setSimplexSteps(new double[] {0.05,0.05,0.05,0.05});
	}

	public int getLengthIndex() {
		return 1; // "Length"
	}
	
//	public double[] makePixels(int length) {
//		double[] p = getParams();
//		return makePixels(p,length);
//	}
	
	/** Makes a symmetric 1-D model intensity distribution function.
	 * @param [1]: the length the distribution extends away from its center.
	 * @param [2]: uniform intensity value away from distribution center.
	 * @param [3]: end intensity value at distal ends of the distribution.
	 * @param n: number of data points to return.  If n is even, n-1 is returned. */
	public Point2D.Double[] makePixels(double[] params, int n) {
		double len = params[1];
		double uni = params[2];
		double pend = params[3];
		if (n % 2 == 0) n--;
		int mid = (n-1)/2; // middle of the distribution
		int x = 0; // position index
		Point2D.Double[] pix = new Point2D.Double [n];
//		double[] pix = new double [n];
//		for (int i=0; i<n; i++) pix[i] = 0;
		pix[mid] = new Point2D.Double(x,BEndInt); // b-end value
		int p = mid+1; // positive index
		int m = mid-1; // negative index
		len--;
		while (len>1) {
			pix[p++] = new Point2D.Double(++x,uni);
			pix[m--] = new Point2D.Double(-x,uni);
			len--;
//			p++;
//			m--;
		}
		pix[p++] = new Point2D.Double(++x,pend*(1-len) + uni*(Math.min(1,(0.5+len))));
		pix[m--] = new Point2D.Double(-x,pend*(1-len) + uni*(Math.min(1,(0.5+len))));
		pix[p++] = new Point2D.Double(++x,pend*(len) + uni*(Math.max((len-0.5),0)));
		pix[m--] = new Point2D.Double(-x,pend*(len) + uni*(Math.max((len-0.5),0)));
		
		while (p<n && m>-1) {
			pix[p++] = new Point2D.Double(++x,0);
			pix[m--] = new Point2D.Double(-x,0);
		}
		return pix;
	}

	public Point2D.Double[] makeDisplay(double[] p) {
		double len = p[1];
		double uni = p[2];
		double end = p[3];
		int n = 1 + ((end!=0)?2:0) + ((uni!=0)?(2*(int)Math.ceil(len)):0);
		Point2D.Double[] pts = new Point2D.Double [n];
		int m = -1;
		int last = n-1;
		if (end!=0) {
			pts[++m] = new Point2D.Double(-len,end+uni);
			pts[last-m] = new Point2D.Double(len,end+uni);
		}

		if (uni!=0) {
			pts[++m] = new Point2D.Double(-len,uni);
			pts[last-m] = new Point2D.Double(len,uni);
			int t = (int)Math.ceil(len-1);
			while (t>0) {
				pts[++m] = new Point2D.Double(-t,uni);
				pts[last-m] = new Point2D.Double(t--,uni);
			}
		}
		pts[++m] = new Point2D.Double(0,BEndInt);
//		for (int i=0; i<pts.length; i++) {
//			System.out.println("UniformPendModel.makeDisplay: i="+i+"; x="+pts[i].x+"; y="+pts[i].y);
//		}
		return pts;
	}

	public Object clone() {
		double[] params = this.getParams();
		boolean[] pStates = this.getParamStates();
		int np = params.length;
		double[] cParams = new double[np];
		boolean[] cPStates = new boolean[np];
		for (int i=0; i<np; i++) {
			cParams[i] = params[i];
			cPStates[i] = pStates[i];
		}
		UniformPendModel c = new UniformPendModel(this.getTitle(), this.getDescript(), cParams,cPStates);
		return (Object) c;
	}
}
