/*  DDecon - Distributed Deconvolution Plugin for ImageJ
    Copyright (C) 2007  Ryan Littlefield

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

import java.awt.geom.*;

public class VariLengthModel extends SModel {
	static final double distalIntensity = 1.0;
	static final int indexProxLength = 2;
	
	public VariLengthModel(String title, String desc, double[] dParams, boolean[] dStates) {
		super(title,desc,dParams,dStates);
		setParamNames(new String[] {"Focus","Distal Len","Prox. Len","Cent. Int","Prox. Int"});
//		setParams(dParams);
//		setParamStates(states);
		setIsBoundLo(new boolean[] {true,true,true,true,true});
		setIsBoundHi(new boolean[] {false,true,true,false,false});
		setBoundsLo(new double[] {0.0,3.0,0.0,0.0,0.0});
		setBoundsHi(new double[] {Double.MAX_VALUE,15.0,1.0,Double.MAX_VALUE,Double.MAX_VALUE});
		setSimplexSteps(new double[] {0.02,0.02,0.02,0.02,0.02});
	}

	public int getLengthIndex() {
		return 1; // "Distal Len" param
	}
	
//	public double[] makePixels(int length) {
//		double[] p = getParams();
//		return makePixels(p,length);
//	}
	
	/** Makes a symmetric 1-D model intensity distribution function.
	 * @param [1]: the furthest length the distribution extends away from its center.
	 * @param [2]: the length away from the center of the distribution, where the intensity changes.
	 * @param [3]: intensity value at distribution center.
	 * @param [4]: intensity value proximal to the distribution center.
	 * @param n: number of data points to return.  If n is even, n-1 is returned. */
	public Point2D.Double[] makePixels(double[] params, int n) {
//		System.out.println("Varilengthmodel.makePixels: n="+n);
		double lenD = params[1];
		double lenP = params[2];
		double intC = params[3];
		double intP = params[4];
		double intD = distalIntensity;
		if (n % 2 == 0) n--;
		int mid = (n-1)/2; // middle of the distribution
		int x = 0; // pixel position index
		Point2D.Double[] pix = new Point2D.Double [n];
//		double[] pix = new double [n];
//		for (int i=0; i<n; i++) pix[i] = 0;
		pix[mid] = new Point2D.Double(x,intC); // center intensity
		int p = mid+1; // positive index
		int m = mid-1; // negative index
		lenD--;
		lenP--;
		while (lenP>0.5) {
			pix[p++] = new Point2D.Double(++x,intP);
			pix[m--] = new Point2D.Double(-x,intP);
			lenD--;
			lenP--;
//			p++;
//			m--;
		}
		pix[p++] = new Point2D.Double(++x,intP*(lenP+0.5) + intD*(0.5-lenP));
		pix[m--] = new Point2D.Double(-x,intP*(lenP+0.5) + intD*(0.5-lenP));
		lenD--;
//		p++;
//		m--;
		while (lenD>0.5) {
			pix[p++] = new Point2D.Double(++x,intD);
			pix[m--] = new Point2D.Double(-x,intD);
			lenD--;
//			p++;
//			m--;
		}
		pix[p++] = new Point2D.Double(++x,intD*(0.5+lenD));
		pix[m--] = new Point2D.Double(-x,intD*(0.5+lenD));

		while (p<n && m>-1) {
			pix[p++] = new Point2D.Double(++x,0);
			pix[m--] = new Point2D.Double(-x,0);
		}
		return pix;
	}

	public Point2D.Double[] makeDisplay(double[] p) {
//		double[] p = getParams();
		double lenD = p[1];
		double lenP = p[2];
		double intC = p[3];
		double intP = p[4];
		double intD = distalIntensity;
		int a = (int)Math.ceil(lenP);
		int b = (int)Math.ceil(lenD);
		int n = 1; // point for center of distribution
		n += (intP!=0)?(2*a):0; // points for proximal region
		n += (intD!=0)?(2*(b-a+2)):0; // points for distal region
		
		Point2D.Double[] pts = new Point2D.Double [n];
		int m = 0;
		int last = n-1;
		if (intD!=0) {
			pts[m] = new Point2D.Double(-lenD,intD);
			pts[last-m] = new Point2D.Double(lenD,intD);
			m++;
//			int t = b-a;
			while (b>=a) {
//			while (t>0) {
				pts[m] = new Point2D.Double(-(b-1),intD);
				pts[last-m] = new Point2D.Double(b-1,intD);
				b--;
//				t--;
				m++;
			}
			pts[m] = new Point2D.Double(-lenP,intD);
			pts[last-m] = new Point2D.Double(lenP,intD);
			m++;
		}

		if (intP!=0) {
			pts[m] = new Point2D.Double(-lenP,intP);
			pts[last-m] = new Point2D.Double(lenP,intP);
			m++;
			int t = a-1;
			while (t>0) {
				pts[m] = new Point2D.Double(-t,intP);
				pts[last-m] = new Point2D.Double(t,intP);
				t--;
				m++;
			}
		}
		pts[m] = new Point2D.Double(0,intC);
		return pts;
	}

	public double[] encodeParamValues (double[] params, boolean[] isFixed, double[] boundsLo, double[] boundsHi) {
//		double[] params = getParams();
//		boolean[] isFixed = getParamStates();
		boolean[] isBoundedLo = getIsBoundLo();
		boolean[] isBoundedHi = getIsBoundHi();
//		double[] boundsLo = getBoundsLo();
//		double[] boundsHi = getBoundsHi();
//		for (int i=0; i<isFixed.length; i++) System.out.println("VariLengthModel.encodeParamValues: isFixed["+i+"]="+isFixed[i]);
		int free = 0;
//		System.out.println("VariLengthModel.encodeParamValues: params.length="+params.length);
		for (int i=0; i<params.length; i++) if (!isFixed[i]) free++;
		double [] eParams = new double [free];
		int index = 0;
		for (int i=0; i<params.length; i++) {
//			System.out.println("VariLengthModel.encodeParamValues: i="+i);
			if (!isFixed[i]) {
				if (i == indexProxLength) {
					// first reparameterize Proximal length to fractional value
					// 1 < prox.length < dist.length-1
					params[i] = (params[i]-1)/(params[1]-2);
				}
				if (isBoundedLo[i] && isBoundedHi[i]) {
					// High and low bounds have both been set, encode periodic boundary conditions
					double center = (boundsHi[i] + boundsLo[i])/2;
					double radius = (boundsHi[i] - boundsLo[i])/2;
					eParams[index] = Math.asin((params[i]-center)/radius);
//					System.out.println("VariLengthModel.encodeParamValues: eParams[index]="+eParams[index]+"; index="+index);
				}
				else if (isBoundedLo[i] && !isBoundedHi[i]) {
					// Low bounds only has been set, encode square root of dist. above low boundary
					eParams[index] = Math.sqrt(params[i]-boundsLo[i]);
//					System.out.println("VariLengthModel.encodeParamValues: eParams[index]="+eParams[index]+"; index="+index);
				}
				else if (!isBoundedLo[i] && isBoundedHi[i]) {
					// High bounds only has been set, encode square root of dist. below high boundary
					eParams[index] = Math.sqrt(boundsHi[i]-params[i]);
//					System.out.println("VariLengthModel.encodeParamValues: eParams[index]="+eParams[index]+"; index="+index);
				}
				else {
					// No bounds set, pass the parameter
					eParams[index] = params[i];
//					System.out.println("VariLengthModel.encodeParamValues: eParams[index]="+eParams[index]+"; index="+index);
				}
				index++;
//				System.out.println("VariLengthModel.encodeParamValues: index="+index);
			}
		}
//		System.out.println("VariLengthModel.encodeParamValues: eParams[0]="+eParams[0]);
		return eParams;
	}
	
	public double[] decodeParamValues(double[] eParams, boolean[] isFixed, double[] boundsLo, double[] boundsHi) {
		double[] params = getParams();
//		boolean[] isFixed = getParamStates();
		boolean[] isBoundedLo = getIsBoundLo();
		boolean[] isBoundedHi = getIsBoundHi();
//		double[] boundsLo = getBoundsLo();
//		double[] boundsHi = getBoundsHi();
		int np = params.length;
		double[] dParams = new double [np];
		for (int i=0; i<np; i++) {
//			System.out.println("VariLengthModel.decodeParams: i="+i);
			if (!isFixed[i]) {
				if (isBoundedLo[i] && isBoundedHi[i]) {
//					System.out.println("VariLengthModel.decodeParams: bHi[i]="+boundsHi[i]+"; bLo[i]="+boundsLo[i]);
					// High and low bounds have both been set, decode periodic boundary conditions
					double center = (boundsHi[i] + boundsLo[i])/2;
					double radius = (boundsHi[i] - boundsLo[i])/2;
//					System.out.println("VariLengthModel.decodeParamValues: eParams[i]="+eParams[i]+"; i="+i);
					dParams[i] = Math.sin(eParams[i])*radius + center;
//					System.out.println("VariLengthModel.decodeParamValues: dP="+dParams[i]);
				}
				else if (isBoundedLo[i] && !isBoundedHi[i]) {
					// Low bounds only has been set, decode square root of dist. above low boundary
					dParams[i] = eParams[i]*eParams[i] + boundsLo[i];
//					System.out.println("VariLengthModel.decodeParamValues: dP="+dParams[i]);
				}
				else if (!isBoundedLo[i] && isBoundedHi[i]) {
					// High bounds only has been set, decode square root of dist. below high boundary
					dParams[i] = boundsHi[i] - eParams[i]*eParams[i];
//					System.out.println("VariLengthModel.decodeParamValues: dP="+dParams[i]);
				}
				else {
					// No bounds set, pass the parameter
					dParams[i] = eParams[i];
//					System.out.println("VariLengthModel.decodeParamValues: dP="+dParams[i]);
				}
				if (i == indexProxLength) {
//					System.out.println("VariLengthModel.decodeParams: i="+i);
//					System.out.println("VariLengthModel.decodeParams: dParams[i]="+dParams[i]);
//					System.out.println("VariLengthModel.decodeParams: lengthIndex="+getLengthIndex());
//					System.out.println("VariLengthModel.decodeParams: dParams[index]="+dParams[getLengthIndex()]);
					// reparameterize proximal length to pixel value
					dParams[i] = dParams[i]*(dParams[getLengthIndex()]-2)+1;
				}
			}
			else dParams[i] = params[i];
		}
//		for (int i=0; i<dParams.length; i++) {
//			System.out.println("VariLengthModel.decodeParamValues: i="+i+"; dP="+dParams[i]);
//		}
		return dParams;
	}
	
	public Object clone() {
		VariLengthModel c = new VariLengthModel(this.getTitle(), this.getDescript(), this.getParams(),this.getParamStates());
		return (Object) c;
	}
}
